from ..layers import (
    AdaptiveFeatureNorm,
    BatchSpectralLoss,
    BNMLoss,
    DiversityLoss,
    EntropyLoss,
    MCCLoss,
)
from ..utils import common_functions as c_f
from .base import BaseHook
from .features import FeaturesAndLogitsHook, FeaturesHook


class BaseLossHook(BaseHook):
    def __init__(self, loss_fn, loss_name, layer, f_hook=None, domains=None, **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = loss_fn
        self.loss_name = loss_name
        self.layer = layer
        self.domains = c_f.default(domains, ["src", "target"])
        self.f_hook = c_f.default(
            f_hook, FeaturesAndLogitsHook, {"domains": self.domains}
        )
        self.regex = [f"^{k}" for k in self.domains]

    def call(self, inputs, losses):
        outputs = self.f_hook(inputs, losses)[0]
        strs = c_f.filter(self.f_hook.out_keys, f"_{self.layer}$", self.regex)
        features = c_f.extract([outputs, inputs], strs)
        loss = sum(self.loss_fn(f) for f in features)
        return outputs, {self.loss_name: loss}

    def _loss_keys(self):
        return [self.loss_name]

    def _out_keys(self):
        return self.f_hook.out_keys

    def extra_repr(self):
        return c_f.extra_repr(self, ["layer", "domains"])


class BSPHook(BaseLossHook):
    def __init__(self, loss_fn=None, domains=None, **kwargs):
        loss_fn = c_f.default(loss_fn, BatchSpectralLoss, {"k": 1})
        domains = c_f.default(domains, ["src", "target"])
        f_hook = FeaturesHook(domains=domains)
        super().__init__(
            loss_fn=loss_fn,
            loss_name="bsp_loss",
            domains=domains,
            layer="features",
            f_hook=f_hook,
            **kwargs,
        )


class BNMHook(BaseLossHook):
    def __init__(self, loss_fn=None, domains=None, **kwargs):
        loss_fn = c_f.default(loss_fn, BNMLoss, {})
        domains = c_f.default(domains, ["target"])
        f_hook = FeaturesAndLogitsHook(domains=domains)
        super().__init__(
            loss_fn=loss_fn,
            loss_name="bnm_loss",
            domains=domains,
            layer="logits",
            f_hook=f_hook,
            **kwargs,
        )


class TargetEntropyHook(BaseLossHook):
    def __init__(self, loss_fn=None, domains=None, **kwargs):
        loss_fn = c_f.default(loss_fn, EntropyLoss, {})
        domains = c_f.default(domains, ["target"])
        f_hook = FeaturesAndLogitsHook(domains=domains)
        super().__init__(
            loss_fn=loss_fn,
            loss_name="entropy_loss",
            domains=domains,
            layer="logits",
            f_hook=f_hook,
            **kwargs,
        )


class TargetDiversityHook(BaseLossHook):
    def __init__(self, loss_fn=None, domains=None, **kwargs):
        loss_fn = c_f.default(loss_fn, DiversityLoss, {})
        domains = c_f.default(domains, ["target"])
        f_hook = FeaturesAndLogitsHook(domains=domains)
        super().__init__(
            loss_fn=loss_fn,
            loss_name="diversity_loss",
            domains=domains,
            layer="logits",
            f_hook=f_hook,
            **kwargs,
        )


class MCCHook(BaseLossHook):
    def __init__(self, loss_fn=None, domains=None, **kwargs):
        loss_fn = c_f.default(loss_fn, MCCLoss, {})
        domains = c_f.default(domains, ["target"])
        f_hook = FeaturesAndLogitsHook(domains=domains)
        super().__init__(
            loss_fn=loss_fn,
            loss_name="mcc_loss",
            domains=domains,
            layer="logits",
            f_hook=f_hook,
            **kwargs,
        )


class AFNHook(BaseLossHook):
    def __init__(self, loss_fn=None, domains=None, **kwargs):
        loss_fn = c_f.default(loss_fn, AdaptiveFeatureNorm, {})
        domains = c_f.default(domains, ["src", "target"])
        f_hook = FeaturesHook(domains=domains)
        super().__init__(
            loss_fn=loss_fn,
            loss_name="afn_loss",
            domains=domains,
            layer="features",
            f_hook=f_hook,
            **kwargs,
        )



import torch

from ..utils import common_functions as c_f
from .entropy_loss import get_entropy


class DiversityLoss(torch.nn.Module):
    """
    Encourages predictions to be uniform, batch wise.
    Takes logits (before softmax) as input.

    For example:

    - A tensor with a large loss: ```torch.tensor([[1e4, 0, 0], [1e4, 0, 0], [1e4, 0, 0]])```

    - A tensor with a small loss: ```torch.tensor([[1e4, 0, 0], [0, 1e4, 0], [0, 0, 1e4]])```
    """

    def __init__(self, after_softmax: bool = False):
        """
        Arguments:
            after_softmax: If ```True```, then the rows of the input are assumed to
                already have softmax applied to them.
        """
        super().__init__()
        self.after_softmax = after_softmax

    def forward(self, logits):
        """"""
        if not self.after_softmax:
            logits = torch.softmax(logits, dim=1)
        logits = torch.mean(logits, dim=0, keepdim=True)
        return -torch.mean(get_entropy(logits, after_softmax=True))

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["after_softmax"])



import torch
import torch.nn.functional as F
from pytorch_metric_learning.distances import CosineSimilarity

from ..utils import common_functions as c_f
from .diversity_loss import DiversityLoss
from .entropy_loss import EntropyLoss


def get_probs(mat, mask, y, dist_is_inverted):
    if not dist_is_inverted:
        mat *= -1
    mat = F.softmax(mat, dim=1)
    n, m = mat.shape
    y = y.repeat(n, 1)[mask].view(n, m)

    target_probs = torch.sum(mat * y, dim=1, keepdims=True)
    src_probs = torch.sum(mat * (1 - y), dim=1, keepdims=True)
    return torch.cat([src_probs, target_probs], dim=1)


def get_loss(probs, ent_fn, div_fn, with_ent, with_div):
    loss = 0
    if with_ent:
        loss += -ent_fn(probs)
    if with_div:
        loss += -div_fn(probs)
    return loss


class ISTLoss(torch.nn.Module):
    """
    Implementation of the I_st loss from
    [Information-Theoretical Learning of Discriminative Clusters for Unsupervised Domain Adaptation](https://icml.cc/2012/papers/566.pdf)
    """

    def __init__(self, distance=None, with_ent=True, with_div=True):
        super().__init__()
        self.distance = c_f.default(distance, CosineSimilarity, {})
        if not (with_ent or with_div):
            raise ValueError("At least one of with_ent or with_div must be True")
        self.with_ent = with_ent
        self.with_div = with_div
        self.ent_loss_fn = EntropyLoss(after_softmax=True)
        self.div_loss_fn = DiversityLoss(after_softmax=True)

    def forward(self, x, y):
        """
        Arguments:
            x: source and target features
            y: domain labels, i.e. 0 for source domain, 1 for target domain
        """
        n = x.shape[0]
        if torch.min(y) < 0 or torch.max(y) > 1:
            raise ValueError("y must be in the range 0 and 1")
        if y.shape != torch.Size([n]):
            raise TypeError("y must have shape (N,)")

        mat = self.distance(x)
        # remove self comparisons
        mask = ~torch.eye(n, dtype=torch.bool)
        mat = mat[mask].view(n, n - 1)
        probs = get_probs(mat, mask, y, self.distance.is_inverted)

        return get_loss(
            probs, self.ent_loss_fn, self.div_loss_fn, self.with_ent, self.with_div
        )

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["with_div"])



from collections import defaultdict

import pytorch_lightning as pl
import torch

from ...containers import Optimizers
from .. import utils as f_utils


def set_adapter_optimizers_to_pl(adapter, pl_optimizers):
    if isinstance(adapter.optimizers, Optimizers):
        keys = adapter.optimizers.keys()
        adapter.optimizers = {k: v for k, v in zip(keys, pl_optimizers)}


def single_dataloader_collect(outputs):
    outputs_as_dict = defaultdict(list)
    for x in outputs:
        for k, v in x.items():
            outputs_as_dict[k].append(v)
    return {k: torch.cat(v) for k, v in outputs_as_dict.items()}


def multi_dataloader_collect(outputs):
    return [single_dataloader_collect(x) for x in outputs]


class Lightning(pl.LightningModule):
    """
    Converts an [Adapter](../../adapters/index.md) into a PyTorch
    Lightning module.
    """

    def __init__(self, adapter, validator=None):
        """
        Arguments:
            adapter:
            validator:
        """
        super().__init__()
        self.models = torch.nn.ModuleDict(adapter.models)
        self.misc = torch.nn.ModuleDict(adapter.misc)
        adapter.models = self.models
        adapter.misc = self.misc
        self.validator = validator
        self.adapter = adapter
        self.automatic_optimization = False

    def forward(self, x, domain=None):
        """"""
        return self.adapter.inference(x, domain=domain)

    def training_step(self, batch, batch_idx):
        """"""
        set_adapter_optimizers_to_pl(self.adapter, self.optimizers())
        losses = self.adapter.training_step(
            batch,
            custom_backward=self.manual_backward,
        )
        for k, v in losses.items():
            self.log(k, v)

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        """"""
        return f_utils.collector_step(self, batch, f_utils.create_output_dict)

    def validation_epoch_end(self, outputs):
        """"""
        required_data = self.validator.required_data
        if len(required_data) > 1:
            outputs = multi_dataloader_collect(outputs)
            data = {k: v for k, v in zip(required_data, outputs)}
        else:
            outputs = single_dataloader_collect(outputs)
            data = {required_data[0]: outputs}
        score = self.validator(**data)
        self.log("validation_score", score)

    def configure_optimizers(self):
        """"""
        optimizers = list(self.adapter.optimizers.values())
        lr_schedulers = []
        for interval in ["epoch", "step"]:
            for v in self.adapter.lr_schedulers.filter_by_scheduler_type(
                f"per_{interval}"
            ):
                lr_schedulers.append({"lr_scheduler": v, "interval": interval})
        return optimizers, lr_schedulers



from ..utils import common_functions as c_f
from .base_container import BaseContainer


class LRSchedulers(BaseContainer):
    """
    A container for optimizer learning rate schedulers.
    """

    def __init__(self, store, scheduler_types=None, **kwargs):
        """
        Arguments:
            store: See [```BaseContainer```][pytorch_adapt.containers.BaseContainer]
            scheduler_types: A dictionary mapping from
                scheduler type (```"per_step"``` or ```"per_epoch"```)
                to a list of object names. If ```None```, then all
                schedulers are assumed to be ```"per_step"```
            **kwargs: [```BaseContainer```][pytorch_adapt.containers.BaseContainer]
                keyword arguments.
        """
        self.scheduler_types = scheduler_types
        super().__init__(store, **kwargs)

    def _create_with(self, other):
        to_be_deleted = []
        for k, v in self.items():
            try:
                class_ref, kwargs = v
            except TypeError:
                continue
            optimizer = other[k]
            if not c_f.is_optimizer(optimizer):
                to_be_deleted.append(k)
            else:
                self[k] = class_ref(optimizer, **kwargs)

        for k in to_be_deleted:
            del self[k]

    def step(self, scheduler_type: str):
        """
        Step the lr schedulers of the specified type.
        Arguments:
            scheduler_type: ```"per_step"``` or ```"per_epoch"```
        """
        for v in self.filter_by_scheduler_type(scheduler_type):
            v.step()

    def filter_by_scheduler_type(self, x):
        if self.scheduler_types is not None:
            return [v for k, v in self.items() if k in self.scheduler_types[x]]
        elif x == "per_step":
            return self.values()
        elif x == "per_epoch":
            return []
        else:
            raise ValueError(
                f"scheduler types are 'per_step' or 'per_epoch', but input is '{x}'"
            )

    def merge(self, other):
        super().merge(other)
        if other.scheduler_types is not None:
            if self.scheduler_types is not None:
                for k, v in other.scheduler_types.items():
                    curr_list = self.scheduler_types[k]
                    curr_list.extend(v)
                    self.scheduler_types[k] = list(set(curr_list))
            else:
                self.scheduler_types = other.scheduler_types



from .checkpoint_utils import CheckpointFnCreator
from .ignite import Ignite
from .ignite_multilabel_classification import IgniteMultiLabelClassification
from .ignite_preds_as_features import IgnitePredsAsFeatures
from .ignite_val_hook_wrapper import IgniteValHookWrapper



import torch

from ..layers import UniformDistributionLoss
from .gan import GANHook


class DomainConfusionHook(GANHook):
    """
    Implementation of
    [Simultaneous Deep Transfer Across Domains and Tasks](https://arxiv.org/abs/1510.02192)
    """

    def __init__(self, **kwargs):
        super().__init__(
            disc_domain_loss_fn=torch.nn.CrossEntropyLoss(reduction="none"),
            gen_domain_loss_fn=UniformDistributionLoss(),
            **kwargs,
        )



from collections import defaultdict

import torch
from pytorch_metric_learning.utils.accuracy_calculator import (
    AccuracyCalculator,
    maybe_get_avg_of_avgs,
    try_getting_not_lone_labels,
    zero_accuracy,
)

from .base_validator import BaseValidator


# temporary
def mean_average_precision(
    knn_labels,
    gt_labels,
    embeddings_come_from_same_source,
    avg_of_avgs,
    return_per_class,
    label_comparison_fn,
    relevance_mask=None,
    at_r=False,
):
    device = gt_labels.device
    num_samples, num_k = knn_labels.shape[:2]
    relevance_mask = (
        torch.ones((num_samples, num_k), dtype=torch.bool, device=device)
        if relevance_mask is None
        else relevance_mask
    )
    is_same_label = label_comparison_fn(gt_labels, knn_labels)
    equality = is_same_label * relevance_mask
    cumulative_correct = torch.cumsum(equality, dim=1)
    k_idx = torch.arange(1, num_k + 1, device=device).repeat(num_samples, 1)
    precision_at_ks = (cumulative_correct * equality).type(torch.float64) / k_idx
    summed_precision_per_row = torch.sum(precision_at_ks * relevance_mask, dim=1)
    if at_r:
        max_possible_matches_per_row = torch.sum(relevance_mask, dim=1)
    else:
        max_possible_matches_per_row = torch.sum(equality, dim=1)
        max_possible_matches_per_row[max_possible_matches_per_row == 0] = 1
    accuracy_per_sample = summed_precision_per_row / max_possible_matches_per_row
    return accuracy_per_sample


# temporary, eventaully move to PML
class BatchedAccuracyCalculator(AccuracyCalculator):
    def __init__(self, batch_size, **kwargs):
        super().__init__(**kwargs)
        self.batch_size = batch_size

    def _get_accuracy(self, function_dict, **kwargs):
        output = defaultdict(list)

        for s in range(0, len(kwargs["knn_labels"]), self.batch_size):
            curr_kwargs = {
                k: kwargs[k][s : s + self.batch_size]
                for k in ["knn_labels", "query_labels", "not_lone_query_mask"]
            }
            curr_kwargs["embeddings_come_from_same_source"] = kwargs[
                "embeddings_come_from_same_source"
            ]
            curr_kwargs["label_counts"] = kwargs["label_counts"]
            curr_results = {k: v(**curr_kwargs) for k, v in function_dict.items()}
            for k, v in curr_results.items():
                output[k].append(v)

        for k, v in output.items():
            output[k] = maybe_get_avg_of_avgs(
                torch.cat(v, dim=0),
                kwargs["query_labels"],
                self.avg_of_avgs,
                self.return_per_class,
            )
        return output

    def calculate_mean_average_precision(
        self,
        knn_labels,
        query_labels,
        not_lone_query_mask,
        embeddings_come_from_same_source,
        label_counts,
        **kwargs,
    ):
        knn_labels, query_labels = try_getting_not_lone_labels(
            knn_labels, query_labels, not_lone_query_mask
        )
        if knn_labels is None:
            return zero_accuracy(label_counts[0], self.return_per_class)

        return mean_average_precision(
            knn_labels,
            query_labels[:, None],
            embeddings_come_from_same_source,
            self.avg_of_avgs,
            self.return_per_class,
            self.label_comparison_fn,
        )


class KNNValidator(BaseValidator):
    def __init__(
        self,
        layer="features",
        metric="precision_at_1",
        k=1,
        knn_func=None,
        kmeans_func=None,
        batch_size=None,
        device=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        acc_kwargs = {
            "k": k,
            "include": (metric,),
            "avg_of_avgs": True,
            "knn_func": knn_func,
            "kmeans_func": kmeans_func,
            "device": device,
        }
        if batch_size is not None:
            acc_kwargs["batch_size"] = batch_size
            acc_fn = BatchedAccuracyCalculator
        else:
            acc_fn = AccuracyCalculator
        self.acc_fn = acc_fn(**acc_kwargs)
        self.layer = layer
        self.metric = metric

    def compute_score(self, src_train, target_train):
        features = torch.cat([src_train[self.layer], target_train[self.layer]], dim=0)
        labels = torch.cat([src_train["domain"], target_train["domain"]], dim=0)
        accuracies = self.acc_fn.get_accuracy(features, features, labels, labels, True)
        return -accuracies[self.metric]



from ..layers import EntropyLoss
from .simple_loss_validator import SimpleLossValidator


class EntropyValidator(SimpleLossValidator):
    """
    Returns the negative of the
    [entropy][pytorch_adapt.layers.entropy_loss.EntropyLoss]
    of all logits.
    """

    @property
    def loss_fn(self):
        return EntropyLoss(after_softmax=self.layer == "preds")



from .accuracy_validator import AccuracyValidator


class ErrorValidator(AccuracyValidator):
    """
    Returns ```-(1-accuracy)```
    """

    def compute_score(self, src_val):
        return -(1 - super().compute_score(src_val))



import torch.nn as nn


class Classifier(nn.Module):
    """
    A 3-layer MLP for classification.
    """

    def __init__(self, num_classes, in_size=2048, h=1024):
        """
        Arguments:
            num_classes: size of the output
            in_size: size of the input
            h: hidden layer size
        """
        super().__init__()
        self.h = h
        self.net = nn.Sequential(
            nn.Linear(in_size, h),
            nn.ReLU(),
            nn.Dropout(),
            nn.Linear(h, h // 2),
            nn.ReLU(),
            nn.Dropout(),
            nn.Linear(h // 2, num_classes),
        )

    def forward(self, x):
        """"""
        return self.net(x)



from .diversity_validator import DiversityValidator
from .entropy_validator import EntropyValidator
from .multiple_validators import MultipleValidators


class IMValidator(MultipleValidators):
    """
    The sum of [EntropyValidator][pytorch_adapt.validators.EntropyValidator]
    and [DiversityValidator][pytorch_adapt.validators.DiversityValidator]
    """

    def __init__(self, weights=None, **kwargs):
        layer = kwargs.pop("layer", None)
        inner_kwargs = {} if not layer else {"layer": layer}
        validators = {
            "entropy": EntropyValidator(**inner_kwargs),
            "diversity": DiversityValidator(**inner_kwargs),
        }
        super().__init__(validators=validators, weights=weights, **kwargs)



import torch

from ..layers import ISTLoss
from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .features import FeaturesHook


class ISTLossHook(BaseWrapperHook):
    def __init__(self, distance=None, with_div=True, **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = ISTLoss(distance=distance, with_div=with_div)
        self.hook = FeaturesHook()

    def call(self, inputs, losses):
        outputs = self.hook(inputs, losses)[0]
        [src_features, target_features] = c_f.extract(
            [outputs, inputs],
            c_f.filter(self.hook.out_keys, "_features$", ["^src", "^target"]),
        )
        features = torch.cat([src_features, target_features], dim=0)
        [src_domain, target_domain] = c_f.extract(
            inputs, ["src_domain", "target_domain"]
        )
        domain = torch.cat([src_domain, target_domain], dim=0)
        loss = self.loss_fn(features, domain)
        return outputs, {"ist_loss": loss}

    def _loss_keys(self):
        return ["ist_loss"]



from pytorch_metric_learning.utils import common_functions as pml_cf

from .ignite_empty_logger import IgniteEmptyLogger


class BasicLossLogger(IgniteEmptyLogger):
    def __init__(self):
        from record_keeper import RecordKeeper, RecordWriter
        from record_keeper import utils as rk_utils

        # Define in here so that record_keeper import stays in init
        class LossKeeper(RecordWriter):
            def __init__(self):
                self.records = self.get_empty_nested_dict()

            def append(self, group_name, series_name, input_val, iteration):
                curr_dict = self.records[group_name]
                append_this = rk_utils.convert_to_scalar(input_val)
                if series_name not in curr_dict:
                    curr_dict[series_name] = []
                curr_dict[series_name].append(append_this)

        self.record_keeper = RecordKeeper(
            record_writer=LossKeeper(),
            attributes_to_search_for=pml_cf.list_of_recordable_attributes_list_names(),
        )

    def add_training(self, _):
        def fn(engine):
            self.record_keeper.update_records(
                engine.state.output, engine.state.iteration
            )

        return fn

    def get_losses(self):
        rw = self.record_keeper.record_writer
        output = rw.records
        rw.records = rw.get_empty_nested_dict()
        return output



from torch.utils.data import Dataset

from ..utils import common_functions as c_f


class DomainDataset(Dataset):
    def __init__(self, dataset: Dataset, domain: int):
        self.dataset = dataset
        self.domain = domain

    def __len__(self):
        return len(self.dataset)

    def __repr__(self):
        return c_f.nice_repr(
            self, c_f.extra_repr(self, ["domain"]), {"dataset": self.dataset}
        )



import copy

from ..containers import KeyEnforcer, MultipleContainers, Optimizers
from ..hooks import ADDAHook
from ..inference import adda_fn
from ..utils import common_functions as c_f
from .base_adapter import BaseAdapter
from .utils import default_optimizer_tuple, with_opt


class ADDA(BaseAdapter):
    """
    Wraps [ADDAHook][pytorch_adapt.hooks.ADDAHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "D"]```|
    |optimizers|```["D", "T"]```|

    The target model ("T") is created during initialization by deep-copying the G model.
    """

    def __init__(self, *args, inference_fn=None, **kwargs):
        """
        Arguments:
            inference_fn: Default is [adda_fn][pytorch_adapt.inference.adda_fn]
        """
        inference_fn = c_f.default(inference_fn, adda_fn)
        super().__init__(*args, inference_fn=inference_fn, **kwargs)

    def get_default_containers(self) -> MultipleContainers:
        """
        Returns:
            The default set of containers. This will use the
                [default optimizer][pytorch_adapt.adapters.utils.default_optimizer_tuple]
                for the T and D models.
        """
        optimizers = Optimizers(default_optimizer_tuple(), keys=["T", "D"])
        return MultipleContainers(optimizers=optimizers)

    def get_key_enforcer(self) -> KeyEnforcer:
        return KeyEnforcer(
            models=["G", "C", "D", "T"],
            optimizers=["D", "T"],
        )

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(
            d_opts=with_opt(["D"]), g_opts=with_opt(["T"]), **hook_kwargs
        )

    def init_containers_and_check_keys(self, containers):
        containers["models"]["T"] = copy.deepcopy(containers["models"]["G"])
        super().init_containers_and_check_keys(containers)

    @property
    def hook_cls(self):
        return ADDAHook



import sys

import setuptools

sys.path.insert(0, "src")
import pytorch_adapt

with open("README.md", "r") as fh:
    long_description = fh.read()


extras_require_detection = ["albumentations >= 1.2.1"]
extras_require_ignite = ["pytorch-ignite == 0.4.9"]
extras_require_lightning = ["pytorch-lightning"]
extras_require_record_keeper = ["record-keeper >= 0.9.32", "tensorboard"]
extras_require_timm = ["timm"]
extras_require_docs = [
    "mkdocs-material",
    "mkdocstrings[python]",
    "griffe",
    "mkdocs-gen-files",
    "mkdocs-section-index",
    "mkdocs-literate-nav",
]
extras_require_dev = ["black", "isort", "nbqa", "flake8"]

setuptools.setup(
    name="pytorch-adapt",
    version=pytorch_adapt.__version__,
    author="Kevin Musgrave",
    description="Domain adaptation made easy. Fully featured, modular, and customizable.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KevinMusgrave/pytorch-adapt",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.0",
    install_requires=[
        "numpy",
        "torch",
        "torchvision",
        "torchmetrics == 0.9.3",
        "pytorch-metric-learning >= 1.6.3",
    ],
    extras_require={
        "detection": extras_require_detection,
        "ignite": extras_require_ignite,
        "lightning": extras_require_lightning,
        "record-keeper": extras_require_record_keeper,
        "timm": extras_require_timm,
        "docs": extras_require_docs,
        "dev": extras_require_dev,
    },
)



from .classifier import Classifier
from .mnist import MNISTFeatures
from .utils import download_weights


def mnistG(pretrained=False, progress=True, **kwargs):
    """
    Returns:
        An [```MNISTFeatures```][pytorch_adapt.models.MNISTFeatures] model
            trained on the MNIST dataset, if ```pretrained == True```.
    """
    model = MNISTFeatures()
    url = "https://cornell.box.com/shared/static/tdx0ts24e273j7mf3r2ox7a12xh4fdfy"
    h = "68ee79452f1d5301be2329dfa542ac6fa18de99e09d6540838606d9d700b09c8"
    file_name = f"mnistG-{h[:8]}.pth"
    return download_weights(
        model, url, pretrained, progress=progress, file_name=file_name, **kwargs
    )


def mnistC(
    num_classes=10, in_size=1200, h=256, pretrained=False, progress=True, **kwargs
):
    """
    Returns:
        A [```Classifier```][pytorch_adapt.models.Classifier] model
            trained on the MNIST dataset, if ```pretrained == True```.
    """
    model = Classifier(num_classes=num_classes, in_size=in_size, h=h)
    url = "https://cornell.box.com/shared/static/j4zrogronmievq1csulrkai7zjm27gcq"
    h = "ac7b5a13df2ef3522b6550a147eb44dde8ff4fead3ddedc540d9fe63c9d597c1"
    file_name = f"mnistC-{h[:8]}.pth"
    return download_weights(
        model, url, pretrained, progress=progress, file_name=file_name, **kwargs
    )


def resnet50(pretrained=False, progress=True, **kwargs):
    import timm

    model = timm.create_model("resnet50", pretrained=False, num_classes=0)
    url = "https://cornell.box.com/shared/static/1oxb5xk5dq3od1d3gprigznxmqb3wgr1"
    h = "a567ecd6ea5addf29ccfa8bf706be78a35adff7cd5cf5a3a99d89b19807454ae"
    file_name = f"resnet50MusgraveUDA-{h[:8]}.pth"
    return download_weights(
        model, url, pretrained, progress=progress, file_name=file_name, **kwargs
    )


def office31G(*args, **kwargs):
    """
    Returns:
        A ResNet50 model trained on ImageNet, if ```pretrained == True```.
    """
    # G was frozen during finetuning
    return resnet50(*args, **kwargs)


def office31C(
    domain=None,
    num_classes=31,
    in_size=2048,
    h=256,
    pretrained=False,
    progress=True,
    **kwargs,
):
    """
    Returns:
        A [```Classifier```][pytorch_adapt.models.Classifier] model
            trained on the specified ```domain``` of the [Office31][pytorch_adapt.datasets.Office31]
            dataset, if ```pretrained == True```. For example

            ```python
            model = office31(domain="amazon", pretrained=True)
            ```
    """
    if (pretrained and not domain) or (not pretrained and domain):
        raise ValueError("if pretrained, domain must be specified, and vice versa")

    model = Classifier(num_classes=num_classes, in_size=in_size, h=h)
    if not pretrained:
        return model
    url = {
        "amazon": "https://cornell.box.com/shared/static/6h165jqlxcpo16jbs3a7vpvslb6u9vaq",
        "dslr": "https://cornell.box.com/shared/static/t97sedzf4wrto3yfvr8hxivyblqkljiq",
        "webcam": "https://cornell.box.com/shared/static/zuv7be39v8bijwggrvfzlyw1h0pfwrb4",
    }[domain]
    h = {
        "amazon": "6e2fb6f392538172515c2c673a8b3ead7aad8b88b44aad6468c7e9b11761b667",
        "dslr": "fc0acd7a71eb5f12d4af619e5c63bcc42e5a23441bbd105fe0f7a37c26f37d80",
        "webcam": "b2bb55978380fa9ca6452cba30e0ac2a19b7166d8348bcc1554fdabd185e4cdd",
    }[domain]
    file_name = f"office31C{domain}-{h[:8]}.pth"
    return download_weights(
        model, url, pretrained, progress=progress, file_name=file_name, **kwargs
    )


def officehomeG(*args, **kwargs):
    """
    Returns:
        A ResNet50 model trained on ImageNet, if ```pretrained == True```.
    """
    # G was frozen during finetuning
    return resnet50(*args, **kwargs)


def officehomeC(
    domain=None,
    num_classes=65,
    in_size=2048,
    h=256,
    pretrained=False,
    progress=True,
    **kwargs,
):
    """
    Returns:
        A [```Classifier```][pytorch_adapt.models.Classifier] model
            trained on the specified ```domain``` of the [OfficeHome][pytorch_adapt.datasets.OfficeHome]
            dataset, if ```pretrained == True```. For example

            ```python
            model = officehomeC(domain="art", pretrained=True)
            ```
    """
    if (pretrained and not domain) or (not pretrained and domain):
        raise ValueError("if pretrained, domain must be specified, and vice versa")

    model = Classifier(num_classes=num_classes, in_size=in_size, h=h)
    if not pretrained:
        return model
    url = {
        "art": "https://cornell.box.com/shared/static/wxg7v32e2m0jcmq53amhdipty9veb2xx",
        "clipart": "https://cornell.box.com/shared/static/4dhwhj6fkzg9lfgu0mfskt2kby8mznez",
        "product": "https://cornell.box.com/shared/static/r6f3ltgve5g2lrcdtoykj84rlyzqs6ga",
        "real": "https://cornell.box.com/shared/static/1lf1foq65m77pdpc50isdgsc8k71ei29",
    }[domain]
    h = {
        "art": "8db546ff250d2b54899f92e482e80a68411cbe525134d429987b57d3b0571e4b",
        "clipart": "8e145cc6d2df3ff428aeafa43066bbde97d56e9f844b34408bdca74125e62590",
        "product": "472ff36fdf13ec6c1fa1236d1d0800e2a5cf2e3d366b6b63ff5807dff6a761d8",
        "real": "f0c8d6e941d4f488ff2438eb5cccdc59e78f35961e48f03d2186752e5878c697",
    }[domain]
    file_name = f"officehomeC{domain}-{h[:8]}.pth"
    return download_weights(
        model, url, pretrained, progress=progress, file_name=file_name, **kwargs
    )


def domainnet126G(domain=None, pretrained=False, progress=True, **kwargs):
    """
    Returns:
        A ResNet50 model trained on ImageNet, if `pretrained == True`.
            If `pretrained == True` and `domain` is specified, then it returns a ResNet50 model trained on that domain.
            For example:
            ```python
            model = domainnet126G(domain="sketch", pretrained=True)
            ```
    """
    import timm

    if not domain:
        return timm.create_model("resnet50", pretrained=pretrained, num_classes=0)
    elif domain and pretrained:
        model = timm.create_model("resnet50", pretrained=False, num_classes=0)
    else:
        raise ValueError("if domain is specified, then pretrained must be True")
    url = {
        "clipart": "https://cornell.box.com/shared/static/p89emqp3sggj8t1ir8ojjgnmgu3i1h6e",
        "painting": "https://cornell.box.com/shared/static/dytk5919v2rwmzgbz8y34hwrtwdh7zbw",
        "real": "https://cornell.box.com/shared/static/cnwd1l38zamp2rwfykrvbw05ds9bq5vp",
        "sketch": "https://cornell.box.com/shared/static/2ifhf99o9oi45c7f2xfspvze9fnbxiwr",
    }[domain]
    h = {
        "clipart": "1cc22355555ae8cafc812584fed828030809ac7eab105c63d220503e4b28e208",
        "painting": "cd4c23e3e8c66900b7cb45d22507f31d43e5da96e9a3e491463f2da72ee15aec",
        "real": "683ac0757a222edab6bad922c7658bc0b1be983699ab38a181f78b150c778315",
        "sketch": "b8c871dc736440aa2e143b75bc540d35056e13012b0de0afb785cba95ed797f7",
    }[domain]
    file_name = f"domainnet126G{domain}-{h[:8]}.pth"
    return download_weights(
        model, url, pretrained, progress=progress, file_name=file_name, **kwargs
    )


def domainnet126C(
    domain=None,
    num_classes=126,
    in_size=2048,
    h=256,
    pretrained=False,
    progress=True,
    **kwargs,
):
    """
    Returns:
        A [```Classifier```][pytorch_adapt.models.Classifier] model
            trained on the specified ```domain``` of the [DomainNet126][pytorch_adapt.datasets.DomainNet126]
            dataset, if ```pretrained == True```. For example

            ```python
            model = domainnet126C(domain="sketch", pretrained=True)
            ```
    """
    if (pretrained and not domain) or (not pretrained and domain):
        raise ValueError("if pretrained, domain must be specified, and vice versa")

    model = Classifier(num_classes=num_classes, in_size=in_size, h=h)
    if not pretrained:
        return model
    url = {
        "clipart": "https://cornell.box.com/shared/static/w2bkoubtzusbpy5xrjgtm1g2fky5yrvk",
        "painting": "https://cornell.box.com/shared/static/2hsaja8l7tilnmvil019zz1cfjuqzj99",
        "real": "https://cornell.box.com/shared/static/xjftd2p0tn6m73d5ddgzbpfp5k13psu5",
        "sketch": "https://cornell.box.com/shared/static/sjrqw3c8h4eop7od2998vvn9bor66w12",
    }[domain]
    h = {
        "clipart": "3728c17634360447ede69520df0c02d9037c2d70f4e9eedeb8aba8242f3fb15b",
        "painting": "11af52a20808ae106168493f61c84f2a695895162d45eb0718333eb1c2ac737c",
        "real": "fc9ea8491b486f5c7302baa5db594f372cdb488f279c9f82a42c752c153b55e8",
        "sketch": "313876e52463eca7d3c7c8be125e86374f976f2afca6bd3aaaa75b1547f84bcf",
    }[domain]
    file_name = f"domainnet126C{domain}-{h[:8]}.pth"
    return download_weights(
        model, url, pretrained, progress=progress, file_name=file_name, **kwargs
    )


def voc_multilabelG(*args, **kwargs):
    return resnet50(*args, **kwargs)


def voc_multilabelC(
    domain=None,
    num_classes=20,
    in_size=2048,
    h=256,
    pretrained=False,
    progress=True,
    **kwargs,
):
    if pretrained:
        raise ValueError("pretrained=True not yet supported")

    return Classifier(num_classes=num_classes, in_size=in_size, h=h)



from typing import Dict

import torch

from ..utils.common_functions import check_domain


def default_fn(x, models, **kwargs) -> Dict[str, torch.Tensor]:
    """
    The default inference function for [BaseAdapter][pytorch_adapt.adapters.BaseAdapter].
    """
    features = models["G"](x)
    logits = models["C"](features)
    return {"features": features, "logits": logits}


# adabn features and logits
def adabn_fn(x, domain, models, **kwargs) -> Dict[str, torch.Tensor]:
    """
    [AdaBN][pytorch_adapt.adapters.AdaBN] features and logits.
    Arguments:
        x: The input to the model
        domain: 0 for source domain, 1 for target domain.
    """
    domain = check_domain(domain, keep_len=True)
    features = models["G"](x, domain)
    logits = models["C"](features, domain)
    return {"features": features, "logits": logits}


def adda_fn(
    x, domain, models, get_all: bool = False, **kwargs
) -> Dict[str, torch.Tensor]:
    """
    [ADDA][pytorch_adapt.adapters.ADDA] features and logits.
    Arguments:
        x: The input to the model
        domain: If 0, then ```features = G(x)```.
            Otherwise ```features = T(x)```.
        models: Dictionary of models with keys
            ```["G", "C", "T"]```.
        get_all: If ```True```, then return features
            and logits using both ```G``` and ```T```
            as the feature extractor.
    Returns:
        A dictionary of features and logits.

            - If ```get_all``` is ```False```, then the keys are ```{"features", "logits"}```.

            - If ```get_all``` is ```True```,
            then the keys will be ```{"features", "logits", "other_features", "other_logits"}```,
            where the ```other_``` prefix represents the features and logits obtained
            using ```G``` if ```domain == 1``` and ```T``` if ```domain == 0```.
    """
    domain = check_domain(domain)
    fe = "G" if domain == 0 else "T"
    features = models[fe](x)
    logits = models["C"](features)
    output = {"features": features, "logits": logits}
    if get_all:
        fe = "T" if fe == "G" else "G"
        features = models[fe](x)
        logits = models["C"](features)
        output.update({"other_features": features, "other_logits": logits})
    return output


def adda_with_d(**kwargs) -> Dict[str, torch.Tensor]:
    """
    [ADDA][pytorch_adapt.adapters.ADDA] features, logits, and discriminator logits. See
    [adda_fn][pytorch_adapt.inference.adda_fn] for the input arguments.
    Returns:
        discriminator logits as ```"d_logits"```, in addition to
            everything returned by [adda_fn][pytorch_adapt.inference.adda_fn].
    """
    return with_d(fn=adda_fn, **kwargs)


def adda_full_fn(x, **kwargs) -> Dict[str, torch.Tensor]:
    """
    [ADDA][pytorch_adapt.adapters.ADDA] features, logits, discriminator logits,
    other features, other logits, other discriminator logits.
    See [adda_fn][pytorch_adapt.inference.adda_fn] for the input arguments.
    Returns:
        discriminator logits (```"d_logits"```), "other" discriminator logits (```"other_d_logits"```)
            in addition to everything returned by [adda_fn][pytorch_adapt.inference.adda_fn]
            with ```get_all = True```.
    """
    layer = kwargs.get("layer", "features")
    output = with_d(x=x, fn=adda_fn, get_all=True, **kwargs)
    output2 = d_fn(x=output[f"other_{layer}"], **kwargs)
    output["other_d_logits"] = output2["d_logits"]
    return output


# rtn features and logits
def rtn_fn(x, domain, models, get_all=False, **kwargs) -> Dict[str, torch.Tensor]:
    """
    [RTN][pytorch_adapt.adapters.RTN] features and logits.
    Arguments:
        x: The input to the model
        domain: If 0, ```logits = residual_model(C(G(x)))```.
            Otherwise, ```logits = C(G(x))```.
        models: Dictionary of models with keys
            ```["G", "C", "residual_model"]```.
        get_all: If ```True```, then in addition to the regular outputs,
            it will return the ```residual_model``` logits when
            ```domain == 1``` and the ```C``` logits when ```domain == 0```.
    Returns:
        A dictionary of features and logits.

            - If ```get_all``` is ```False```, then the keys are ```{"features", "logits"}```.

            - If ```get_all``` is ```True```,
            then the keys will be ```{"features", "logits", "other_logits"}```.
    """
    domain = check_domain(domain)
    f_dict = default_fn(x=x, models=models)
    target_logits = f_dict["logits"]
    if get_all or domain == 0:
        src_logits = models["residual_model"](target_logits)
    if domain == 0:
        f_dict["logits"] = src_logits
        if get_all:
            f_dict["other_logits"] = target_logits
    elif get_all and domain == 1:
        f_dict["other_logits"] = src_logits
    return f_dict


def rtn_with_feature_combiner(**kwargs):
    return with_feature_combiner(fn=rtn_fn, **kwargs)


def rtn_full_fn(**kwargs):
    return rtn_with_feature_combiner(get_all=True, **kwargs)


# mcd features and logits
def mcd_fn(x, models, get_all=False, **kwargs):
    """
    Returns:
        Features and logits, where ```logits = sum(C(features))```.
    """
    features = models["G"](x)
    logits_list = models["C"](features)
    logits = sum(logits_list)
    output = {"features": features, "logits": logits}
    if get_all:
        for i, L in enumerate(logits_list):
            output[f"logits{i}"] = L
    return output


# mcd features and logits, and the logits from each classifier
def mcd_full_fn(**kwargs):
    return mcd_fn(get_all=True, **kwargs)


# symnets features and logits
def symnets_fn(x, domain, models, get_all=False, **kwargs):
    """
    Arguments:
        x: The input to the model
        domain: 0 for the source domain, 1 for the target domain.
    Returns:
        Features and logits, where ```logits = C(features)[domain]```.
    """
    domain = check_domain(domain)
    features = models["G"](x)
    logits = models["C"](features)[domain]
    output = {"features": features, "logits": logits}
    if get_all:
        logits = models["C"](features)[int(not domain)]
        output.update({"other_logits": logits})
    return output


def symnets_full_fn(**kwargs):
    return symnets_fn(get_all=True, **kwargs)


# discriminator logits
def d_fn(x, models, **kwargs):
    return {"d_logits": models["D"](x)}


# output of a feature combiner
def feature_combiner_fn(x, y, misc, **kwargs):
    return {"features_logits_combined": misc["feature_combiner"](x, y)}


# output of fn and d_fn
def with_d(x, models, fn, layer="features", **kwargs):
    output = fn(x=x, models=models, layer=layer, **kwargs)
    output2 = d_fn(x=output[layer], models=models, **kwargs)
    return {**output, **output2}


# features, logits, and discriminator logits
def default_with_d(**kwargs):
    return with_d(fn=default_fn, **kwargs)


# features, logits, and discriminator logits, where D takes in the output of C
def default_with_d_logits_layer(**kwargs):
    return with_d(fn=default_fn, layer="logits", **kwargs)


# output of fn and a feature combiner that takes features and logits or preds as input
def with_feature_combiner(x, models, misc, fn, softmax=True, **kwargs):
    output = fn(x=x, models=models, misc=misc, softmax=softmax, **kwargs)
    logits = output["logits"]
    if softmax:
        logits = torch.softmax(logits, dim=1)
    output2 = feature_combiner_fn(x=output["features"], y=logits, misc=misc, **kwargs)
    return {**output, **output2}


# cdan features, logits, discriminator logits, and feature_combiner output
def cdan_full_fn(**kwargs):
    return with_feature_combiner(fn=default_with_d, **kwargs)


# discriminator bridge logits
def d_bridge_fn(x, models, **kwargs):
    [d_logits, d_bridge] = models["D"](x, return_bridge=True)
    return {"d_logits": d_logits, "d_bridge": d_bridge}


# output of fn and d_bridge_fn. D takes in preds
def with_d_bridge(x, models, fn, **kwargs):
    output = fn(x=x, models=models, **kwargs)
    preds = torch.softmax(output["logits"], dim=1)
    output2 = d_bridge_fn(x=preds, models=models, **kwargs)
    return {**output, **output2}


# gvb features, logits, and g_bridge
def gvb_with_g_bridge(x, models, **kwargs):
    features = models["G"](x)
    [logits, bridge] = models["C"](features, return_bridge=True)
    return {"features": features, "logits": logits, "g_bridge": bridge}


# gvb features, logits, g_bridge, and d_bridge
def gvb_full_fn(**kwargs):
    return with_d_bridge(fn=gvb_with_g_bridge, **kwargs)



from typing import List

from ..utils import common_functions as c_f
from .multiple_containers import MultipleContainers


class KeyEnforcer:
    """
    Makes sure containers have the specified keys.
    """

    def __init__(self, **kwargs: List[str]):
        """
        Arguments:
            **kwargs: A mapping from container name to a list of required
                keys for that container.
        """
        self.requirements = kwargs

    def check(self, containers: MultipleContainers):
        """
        Compares the input containers' keys to ```self.requirements```.
        Raises ```KeyError``` if there is a mismatch.
        Arguments:
            containers: The containers to check.
        """
        for k, required_keys in self.requirements.items():
            container_keys = list(containers[k].keys())
            r_c_diff = c_f.list_diff(required_keys, container_keys)
            c_r_diff = c_f.list_diff(container_keys, required_keys)
            error_msg = ""
            if len(r_c_diff) > 0:
                error_msg += (
                    f"The {k} container is missing the following keys: {r_c_diff}. "
                )

            if len(c_r_diff) > 0:
                error_msg += (
                    f"The {k} container has the following unallowed keys: {c_r_diff}."
                )

            if error_msg != "":
                raise KeyError(error_msg)



class HookLogger:
    def __init__(self, name):
        self.name = name
        self.reset()

    def __call__(self, x):
        if self.str:
            self.str += "\n"
        self.str += f"{self.name}: {x}"

    def reset(self):
        self.str = ""



import math

import torch
import torch.nn as nn

from ..utils import common_functions as c_f


# https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/linear.py
def reset_parameters_helper(weight, bias):
    nn.init.kaiming_uniform_(weight, a=math.sqrt(5))
    if bias is not None:
        fan_in, _ = nn.init._calculate_fan_in_and_fan_out(weight)
        bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
        nn.init.uniform_(bias, -bound, bound)


# https://openaccess.thecvf.com/content_CVPR_2020/papers/Lu_Stochastic_Classifiers_for_Unsupervised_Domain_Adaptation_CVPR_2020_paper.pdf
class StochasticLinear(nn.Module):
    """
    Implementation of the stochastic layer from
    [Stochastic Classifiers for Unsupervised Domain Adaptation](https://xiatian-zhu.github.io/papers/LuEtAl_CVPR2020.pdf).
    In ```train()``` mode, it uses random weights and biases
    that are sampled from a learned normal distribution.
    In ```eval()``` mode, the learned mean is used.
    """

    def __init__(self, in_features: int, out_features: int, device=None, dtype=None):
        """
        Arguments:
            in_features: size of each input sample
            out_features: size of each output sample
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight_mean = nn.Parameter(
            torch.empty(in_features, out_features, **factory_kwargs)
        )
        self.weight_sigma = nn.Parameter(
            torch.empty(in_features, out_features, **factory_kwargs)
        )
        self.bias_mean = nn.Parameter(torch.empty(out_features, **factory_kwargs))
        self.bias_sigma = nn.Parameter(torch.empty(out_features, **factory_kwargs))
        self.reset_parameters()

    def reset_parameters(self):
        reset_parameters_helper(self.weight_mean, self.bias_mean)
        reset_parameters_helper(self.weight_sigma, self.bias_sigma)

    def random_sample(self, mean, sigma):
        eps = torch.randn(*sigma.shape, device=sigma.device, dtype=sigma.dtype)
        return mean + (sigma * eps)

    def forward(self, x):
        """"""
        if self.training:
            weight = self.random_sample(self.weight_mean, self.weight_sigma)
            bias = self.random_sample(self.bias_mean, self.bias_sigma)
        else:
            weight = self.weight_mean
            bias = self.bias_mean

        return torch.matmul(x, weight) + bias

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["in_features", "out_features"])



import torch
from pytorch_metric_learning.distances import BatchedDistance, CosineSimilarity

from ..layers import DiversityLoss, EntropyLoss
from ..layers.ist_loss import get_loss, get_probs
from ..utils.common_functions import mask_out_self
from .base_validator import BaseValidator


def get_iter_fn(probs, y, dist_is_inverted):
    def fn(mat, s, *_):
        mat, mask = mask_out_self(mat, s, return_mask=True)
        p = get_probs(mat, mask, y, dist_is_inverted)
        probs.append(p)

    return fn


class ISTValidator(BaseValidator):
    def __init__(
        self, layer="features", batch_size=1024, with_ent=True, with_div=True, **kwargs
    ):
        super().__init__(**kwargs)
        self.layer = layer
        self.dist_fn = BatchedDistance(CosineSimilarity(), batch_size=batch_size)
        self.with_ent = with_ent
        self.with_div = with_div
        self.ent_fn = EntropyLoss(after_softmax=True)
        self.div_fn = DiversityLoss(after_softmax=True)

    def compute_score(self, src_train, target_train):
        features = torch.cat([src_train[self.layer], target_train[self.layer]], dim=0)
        labels = torch.cat([src_train["domain"], target_train["domain"]], dim=0)

        # probs is modified via self.iter_fn
        probs = []
        self.dist_fn.iter_fn = get_iter_fn(
            probs, labels, self.dist_fn.distance.is_inverted
        )
        self.dist_fn(features)
        probs = torch.cat(probs, dim=0)
        if len(probs) != len(features):
            raise ValueError("probs should have same length as features")
        return get_loss(probs, self.ent_fn, self.div_fn, self.with_ent, self.with_div)



from typing import Tuple


class ForwardOnlyValidator:
    """
    This is basically a pass-through function.
    It returns the best score and best epoch
    that is returned by the inner adapter.
    """

    def run(self, adapter, **kwargs) -> Tuple[float, int]:
        """
        Arguments:
            adapter: the framework-wrapped adapter.
            **kwargs: keyword arguments to be passed into adapter.run()
        Returns:
            the best score and best epoch
        """
        if not adapter.validator:
            raise KeyError(
                "An adapter validator is required when using ForwardOnlyValidator"
            )
        return adapter.run(**kwargs)



from collections.abc import MutableMapping
from typing import Any, Dict, Tuple

import torch


def default_optimizer_tuple() -> Tuple[torch.optim.Optimizer, Dict[str, Any]]:
    """
    Returns:
        A tuple to be passed into an [Optimizers][pytorch_adapt.containers.Optimizers]
            container. The tuple specifies an Adam optimizer with lr 0.0001.
    """
    return (torch.optim.Adam, {"lr": 0.0001})


def with_opt(x):
    suffix = "_opt"
    if isinstance(x, str):
        return f"{x}{suffix}"
    if isinstance(x, list):
        return [with_opt(y) for y in x]
    if isinstance(x, MutableMapping):
        return {with_opt(k): v for k, v in x.items()}


def container_names():
    return ["models", "optimizers", "lr_schedulers", "misc"]



from ..layers import MMDBatchedLoss
from ..utils import common_functions as c_f
from .base_validator import BaseValidator


class MMDValidator(BaseValidator):
    def __init__(self, layer="features", batch_size=1024, mmd_kwargs=None, **kwargs):
        super().__init__(**kwargs)
        self.layer = layer
        mmd_kwargs = c_f.default(mmd_kwargs, {})
        self.loss_fn = MMDBatchedLoss(batch_size=batch_size, **mmd_kwargs)

    def compute_score(self, src_train, target_train):
        x = src_train[self.layer]
        y = target_train[self.layer]
        return -self.loss_fn(x, y).item()



import torch
from pytorch_metric_learning.distances import (
    BatchedDistance,
    CosineSimilarity,
    LpDistance,
)
from pytorch_metric_learning.utils.inference import CustomKNN

from .base_validator import BaseValidator


def acc(preds, labels):
    if max(labels) != preds.shape[1] - 1:
        raise ValueError(
            f"Max label {max(labels)} should be equal to preds.shape[1] {preds.shape[1]}"
        )
    preds = torch.argmax(preds, dim=1)
    return (preds == labels).float()


class NearestSourceValidator(BaseValidator):
    def __init__(self, layer="preds", threshold=0, weighted=False, **kwargs):
        super().__init__(**kwargs)
        self.layer = layer
        self.threshold = threshold
        self.weighted = weighted
        self.knn_fn = CustomKNN(CosineSimilarity())

    def compute_score(self, src_val, target_train):
        nearest_src_acc, sims = self.get_nearest_src_acc(src_val, target_train)

        if self.weighted:
            sims = (sims - self.threshold) / (max(sims) - self.threshold)
            sims[sims <= 0] = 0
            nearest_src_acc *= sims
        else:
            nearest_src_acc[sims <= self.threshold] = 0

        return torch.mean(nearest_src_acc).item()

    def get_nearest_src_acc(self, src_val, target_train):
        src_acc = acc(src_val["preds"], src_val["labels"])

        sims, idx = self.knn_fn(
            target_train[self.layer],
            k=1,
            reference=src_val[self.layer],
            embeddings_come_from_same_source=False,
        )
        sims, idx = sims.squeeze(1), idx.squeeze(1)
        nearest_src_acc = src_acc[idx]
        return nearest_src_acc, sims


class NearestSourceL2Validator(NearestSourceValidator):
    def __init__(self, layer="preds", **kwargs):
        super().__init__(layer=layer, threshold=float("inf"), weighted=True, **kwargs)
        dist_fn = LpDistance(normalize_embeddings=False)
        self.knn_fn = CustomKNN(dist_fn)
        self.all_dist_fn = BatchedDistance(dist_fn, batch_size=1024)

    def compute_score(self, src_val, target_train):
        max_dist = [0]

        def iter_fn(mat, *_):
            max_dist[0] = max(max_dist[0], torch.max(mat))

        all_feats = torch.cat([src_val[self.layer], target_train[self.layer]], dim=0)
        self.all_dist_fn.iter_fn = iter_fn
        self.all_dist_fn(all_feats)

        nearest_src_acc, dists = self.get_nearest_src_acc(src_val, target_train)
        dists /= max_dist[0]
        nearest_src_acc *= 1 - dists
        return torch.mean(nearest_src_acc).item()


#



import os
import tarfile
import zipfile

import torch
from PIL import Image
from torchvision.datasets.utils import download_url

from ..utils import common_functions as c_f


class BaseDataset(torch.utils.data.Dataset):
    """
    Base dataset class
    """

    def __init__(self, domain: str):
        super().__init__()
        self.domain = domain

    def __len__(self):
        return len(self.img_paths)

    def __getitem__(self, idx):
        label = self.labels[idx]
        img = Image.open(self.img_paths[idx]).convert("RGB")
        if self.transform is not None:
            img = self.transform(img)
        return img, label

    def __repr__(self):
        extra_repr = f"domain={self.domain}\nlen={str(self.__len__())}"
        return c_f.nice_repr(self, extra_repr, {"transform": self.transform})


class BaseDownloadableDataset(BaseDataset):
    """
    Allows automatic downloading of datasets.
    """

    def __init__(self, root: str, download: bool = False, **kwargs):
        """
        Arguments:
            root: Folder where dataset will be downloaded to.
            download: If True, will download the dataset if it hasn't already been downloaded.
        """
        super().__init__(**kwargs)
        if download:
            try:
                self.set_paths_and_labels(root)
            except (FileNotFoundError, ValueError):
                self.download_dataset(root)
                self.set_paths_and_labels(root)
        else:
            self.set_paths_and_labels(root)

    def set_paths_and_labels(self, root):
        raise NotImplementedError

    def download_dataset(self, root):
        download_url(self.url, root, filename=self.filename, md5=self.md5)
        filepath = os.path.join(root, self.filename)
        decompressor = tarfile.open if tarfile.is_tarfile(filepath) else zipfile.ZipFile
        c_f.LOGGER.info("Extracting")
        with decompressor(filepath, "r") as f:
            f.extractall(path=root, members=c_f.extract_progress(f))



import os


def check_img_paths(img_dir, img_paths, domain):
    for x in img_paths:
        x = os.path.relpath(x, img_dir)
        if not x.startswith(domain):
            raise ValueError(
                f"img_paths contains a path {x} to an image outside of domain {domain}"
            )


def check_length(cls, correct_length):
    x = len(cls)
    if x != correct_length:
        raise ValueError(f"len(self)={x} but should be {correct_length}")


def check_train(train):
    if not isinstance(train, bool):
        raise TypeError("train should be True or False")
    return train


def maybe_download(fn, kwargs):
    original_download = kwargs["download"]
    try:
        kwargs["download"] = False
        fn(**kwargs)
    except (RuntimeError, FileNotFoundError):
        if original_download:
            kwargs["download"] = True
            fn(**kwargs)
        else:
            raise


def num_classes(dataset_name):
    return {
        "mnist": 10,
        "domainnet": 345,
        "domainnet126": 126,
        "office31": 31,
        "officehome": 65,
        "voc_multilabel": 20,
    }[dataset_name]



from ..layers import VATLoss
from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .features import FeaturesAndLogitsHook
from .gan import GANHook
from .losses import TargetEntropyHook
from .utils import ChainHook


class VATHook(BaseWrapperHook):
    """
    Applies the [```VATLoss```][pytorch_adapt.layers.VATLoss].
    """

    def __init__(self, loss_fn=None, **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = c_f.default(loss_fn, VATLoss, {})
        self.hook = FeaturesAndLogitsHook()

    def call(self, inputs, losses):
        outputs = self.hook(inputs, losses)[0]
        [src_imgs, target_imgs, combined_model] = c_f.extract(
            inputs, ["src_imgs", "target_imgs", "combined_model"]
        )
        [src_logits, target_logits] = c_f.extract(
            [outputs, inputs],
            c_f.filter(self.hook.out_keys, "_logits$", ["^src", "^target"]),
        )
        src_vat_loss = self.loss_fn(src_imgs, src_logits, combined_model)
        target_vat_loss = self.loss_fn(target_imgs, target_logits, combined_model)
        return (
            outputs,
            {
                "src_vat_loss": src_vat_loss,
                "target_vat_loss": target_vat_loss,
            },
        )

    def _loss_keys(self):
        return ["src_vat_loss", "target_vat_loss"]


class VATPlusEntropyHook(BaseWrapperHook):
    def __init__(self, vat_loss_fn=None, entropy_loss_fn=None, **kwargs):
        super().__init__(**kwargs)
        hook1 = VATHook(vat_loss_fn)
        hook2 = TargetEntropyHook(entropy_loss_fn)
        self.hook = ChainHook(hook1, hook2)


class VADAHook(GANHook):
    """
    Implementation of VADA from
    [A DIRT-T Approach to Unsupervised Domain Adaptation](https://arxiv.org/abs/1802.08735).
    """

    def __init__(self, vat_loss_fn=None, entropy_loss_fn=None, post_g=None, **kwargs):
        post_g = c_f.default(post_g, [])
        post_g += [VATPlusEntropyHook(vat_loss_fn, entropy_loss_fn)]
        super().__init__(post_g=post_g, **kwargs)



import re
from typing import List

import torch

from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .utils import ChainHook


class BaseFeaturesHook(BaseHook):
    """
    This hook:

    1. Checks to see if specific tensors are in the context
    2. Exits if the tensors are already in the context
    3. Otherwise computes those tensors using the appropriate
        inputs and models, and adds them to the context.
    """

    def __init__(
        self,
        model_name: str,
        in_suffixes: List[str] = None,
        out_suffixes: List[str] = None,
        domains: List[str] = None,
        detach: bool = False,
        **kwargs,
    ):
        """
        Arguments:
            model_name: The name of the model that will
                be used to compute any missing tensors.

            in_suffixes: The suffixes of the names of the inputs
                to the model. For example if:

                - ```domains = ["src", "target"]```
                - ```in_suffixes = ["_imgs_features"]```

                then the model will be given

                - ```["src_imgs_features", "target_imgs_features"]```.

            out_suffixes: The suffixes of the names of the outputs
                of the model. Output suffixes are appended to the input name.
                For example, if

                - ```domains = ["src", "target"]```
                - ```in_suffixes = ["_imgs_features"]```
                - ```out_suffixes = ["_logits"]```

                then the output keys will be

                - ```["src_imgs_features_logits", "target_imgs_features_logits"]```

            domains: The names of the domains to use. If ```None```,
                this defaults to ```["src", "target"]```.

            detach: If ```True```, then the output will be detached
                from the autograd graph. Any output that is detached
                will have ```"_detached"``` appended to its name in the
                context.
        """

        super().__init__(**kwargs)
        self.model_name = model_name
        self.domains = c_f.default(domains, ["src", "target"])
        self.init_detach_mode(detach)
        self.init_suffixes(in_suffixes, out_suffixes)

    def call(self, inputs, losses):
        """"""
        outputs = {}
        for domain in self.domains:
            self.logger(f"Getting {domain}")
            detach = self.check_grad_mode(domain)
            func = self.mode_detached if detach else self.mode_with_grad
            in_keys = c_f.filter(self.in_keys, f"^{domain}")
            func(inputs, outputs, domain, in_keys)

        self.check_outputs_requires_grad(outputs)
        return outputs, {}

    def check_grad_mode(self, domain):
        detach = self.detach[domain]
        if not torch.is_grad_enabled():
            if not detach:
                raise ValueError(
                    f"detach[{domain}] == {detach} but grad is not enabled"
                )
        return detach

    def check_outputs_requires_grad(self, outputs):
        for k, v in outputs.items():
            if k.endswith("detached") and c_f.requires_grad(v, does=True):
                raise TypeError(f"{k} ends with 'detached' but tensor requires grad")
            if not k.endswith("detached") and c_f.requires_grad(v, does=False):
                raise TypeError(
                    f"{k} doesn't end in 'detached' but tensor doesn't require grad"
                )

    def mode_with_grad(self, inputs, outputs, domain, in_keys):
        output_keys = c_f.filter(self._out_keys(), f"^{domain}")
        output_vals = self.get_kwargs(inputs, output_keys)
        self.add_if_new(
            outputs, output_keys, output_vals, inputs, self.model_name, in_keys, domain
        )
        return output_keys, output_vals

    def mode_detached(self, inputs, outputs, domain, in_keys):
        curr_out_keys = c_f.filter(self._out_keys(), f"^{domain}")
        self.try_existing_detachable(inputs, outputs, curr_out_keys)
        remaining_out_keys = [
            k for k in curr_out_keys if k not in set().union(inputs, outputs)
        ]
        if len(remaining_out_keys) > 0:
            output_vals = self.get_kwargs(inputs, remaining_out_keys)
            with torch.no_grad():
                self.add_if_new(
                    outputs,
                    remaining_out_keys,
                    output_vals,
                    inputs,
                    self.model_name,
                    in_keys,
                    domain,
                )

    def add_if_new(
        self, outputs, full_key, output_vals, inputs, model_name, in_keys, domain
    ):
        c_f.add_if_new(
            outputs,
            full_key,
            output_vals,
            inputs,
            model_name,
            in_keys,
            logger=self.logger,
        )

    def create_keys(self, domain, suffix, starting_keys=None, detach=False):
        if starting_keys is None:
            full_keys = [f"{domain}{x}" for x in suffix]
        else:
            if len(starting_keys) > 1:
                starting_keys = self.join_keys(starting_keys)
            if len(suffix) > 1:
                starting_keys = starting_keys * len(suffix)
            full_keys = [f"{k}{x}" for k, x in zip(starting_keys, suffix)]
        if detach:
            full_keys = self.add_detached_string(full_keys)
        return full_keys

    def get_kwargs(self, inputs, keys):
        return [inputs.get(k) for k in keys]

    def try_existing_detachable(self, inputs, outputs, curr_out_keys):
        for k in curr_out_keys:
            if k in inputs or k in outputs:
                continue
            curr_regex = self.detachable_regex[k]
            success = self.try_existing_detachable_in_dict(
                curr_regex, inputs, outputs, k
            )
            if not success:
                self.try_existing_detachable_in_dict(curr_regex, outputs, outputs, k)

    def try_existing_detachable_in_dict(self, regex, in_dict, outputs, new_k):
        for k, v in in_dict.items():
            if regex.search(k) and v is not None:
                outputs[new_k] = v.detach()
                return True
        return False

    def add_detached_string(self, keys):
        # delete existing detached string, then append to the very end
        # for example, if computing detached logits for: src_imgs_features_detached
        # 1. src_imgs_features_detached_logits --> src_imgs_features_logits
        # 2. src_imgs_features_logits --> src_imgs_features_logits_detached
        keys = [k.replace("_detached", "") for k in keys]
        return [f"{k}_detached" for k in keys]

    def join_keys(self, keys):
        return ["_AND_".join(keys)]

    def init_detach_mode(self, detach):
        if isinstance(detach, dict):
            if any(not isinstance(v, bool) for v in detach.values()):
                raise TypeError("if detach is a dict, values must be bools")
            self.detach = detach
        elif isinstance(detach, bool):
            self.detach = {k: detach for k in self.domains}
        else:
            raise TypeError("detach must be a bool or a dict of bools")

    def init_suffixes(self, in_suffixes, out_suffixes):
        self.in_suffixes = in_suffixes
        self.out_suffixes = out_suffixes
        in_keys = []
        for domain in self.domains:
            in_keys.extend(self.create_keys(domain, in_suffixes))
        self.set_in_keys(in_keys)

    def set_in_keys(self, in_keys):
        super().set_in_keys(in_keys)
        self.all_out_keys = []
        for domain in self.domains:
            curr_in_keys = c_f.filter(self.in_keys, f"^{domain}")
            curr_out_keys = self.create_keys(
                domain, self.out_suffixes, curr_in_keys, detach=self.detach[domain]
            )
            self.all_out_keys.extend(curr_out_keys)

        # strings with '_detached' optional and anywhere
        self.detachable_regex = {
            k: re.compile(
                f"^{k.replace('_detached', '').replace('_', '(_detached)?_')}$"
            )
            for k in self.all_out_keys
        }

    def _loss_keys(self):
        """"""
        return []

    def _out_keys(self):
        """"""
        return self.all_out_keys

    def extra_repr(self):
        return c_f.extra_repr(self, ["model_name", "domains", "detach"])


class FeaturesHook(BaseFeaturesHook):
    """
    Default input/output context names:

    - Model: ```"G"```
    - Inputs:
        ```
        ["src_imgs",
        "target_imgs"]
        ```
    - Outputs:
        ```
        ["src_imgs_features",
        "target_imgs_features"]
        ```
    """

    def __init__(
        self,
        model_name="G",
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs"])
        out_suffixes = c_f.default(out_suffixes, ["_features"])
        super().__init__(
            model_name=model_name,
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class LogitsHook(BaseFeaturesHook):
    """
    Default input/output context names:

    - Model: ```"C"```
    - Inputs:
        ```
        ["src_imgs_features",
        "target_imgs_features"]
        ```
    - Outputs:
        ```
        ["src_imgs_features_logits",
        "target_imgs_features_logits"]
        ```
    """

    def __init__(
        self,
        model_name="C",
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs_features"])
        out_suffixes = c_f.default(out_suffixes, ["_logits"])
        super().__init__(
            model_name=model_name,
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class FeaturesChainHook(ChainHook):
    """
    A special [```ChainHook```][pytorch_adapt.hooks.ChainHook]
    for features hooks. It sets each sub-hook's ```in_keys``` using
    the previous sub-hook's ```out_keys```.
    """

    def __init__(
        self,
        *hooks,
        **kwargs,
    ):
        for i in range(len(hooks) - 1):
            hooks[i + 1].set_in_keys(hooks[i].out_keys)
        self.domains = hooks[-1].domains
        super().__init__(*hooks, **kwargs)


class FeaturesAndLogitsHook(FeaturesChainHook):
    """
    Chains together [```FeaturesHook```][pytorch_adapt.hooks.FeaturesHook]
    and [```LogitsHook```][pytorch_adapt.hooks.LogitsHook].
    """

    def __init__(
        self,
        domains: List[str] = ("src", "target"),
        detach_features: bool = False,
        detach_logits: bool = False,
        other_hooks: List[BaseHook] = None,
        **kwargs,
    ):
        """
        Arguments:
            domains: The domains used by both the features and logits hooks.
                If ```None```, it defaults to ```["src", "target"]```
            detach_features: If ```True```, returns features that are
                detached from the autograd graph.
            detach_logits: If ```True```, returns logits that are
                detached from the autograd graph.
            other_hooks: A list of hooks that will be called after
                the features and logits hooks.
        """
        self.domains = domains
        features_hook = FeaturesHook(detach=detach_features, domains=domains)
        logits_hook = LogitsHook(detach=detach_logits, domains=domains)
        other_hooks = c_f.default(other_hooks, [])
        super().__init__(features_hook, logits_hook, *other_hooks, **kwargs)


class FeaturesWithGradAndDetachedHook(BaseWrapperHook):
    """
    Default input/output context names:

    - Model: ```"G"```
    - Inputs:
        ```
        ["src_imgs",
        "target_imgs"]
        ```
    - Outputs:
        ```
        ["src_imgs_features",
        "target_imgs_features",
        "src_imgs_features_detached",
        "target_imgs_features_detached"]
        ```
    """

    def __init__(
        self,
        model_name="G",
        in_suffixes=None,
        out_suffixes=None,
        domains=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        hooks = []
        for detach in [False, True]:
            hooks.append(
                FeaturesHook(
                    model_name=model_name,
                    in_suffixes=in_suffixes,
                    out_suffixes=out_suffixes,
                    domains=domains,
                    detach=detach,
                    **kwargs,
                )
            )
        self.hook = ChainHook(*hooks)


class CombinedFeaturesHook(BaseFeaturesHook):
    """
    Default input/output context names:

    - Model: ```"feature_combiner"```
    - Inputs:
        ```
        ["src_imgs_features",
        "src_imgs_features_logits",
        "target_imgs_features",
        "target_imgs_features_logits"]
        ```
    - Outputs:
        ```
        ["src_imgs_features_AND_src_imgs_features_logits_combined",
        "target_imgs_features_AND_target_imgs_features_logits_combined"]
        ```
    """

    def __init__(
        self,
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(
            in_suffixes, ["_imgs_features", "_imgs_features_logits"]
        )
        out_suffixes = c_f.default(out_suffixes, ["_combined"])
        super().__init__(
            model_name="feature_combiner",
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class DLogitsHook(BaseFeaturesHook):
    """
    Default input/output context names:

    - Model: ```"D"```
    - Inputs:
        ```
        ["src_imgs_features",
        "target_imgs_features"]
        ```
    - Outputs:
        ```
        ["src_imgs_features_dlogits",
        "target_imgs_features_dlogits"]
        ```
    """

    def __init__(
        self,
        model_name="D",
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs_features"])
        out_suffixes = c_f.default(out_suffixes, ["_dlogits"])
        super().__init__(
            model_name=model_name,
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class FrozenModelHook(BaseWrapperHook):
    """
    Sets model to ```eval()``` mode, and does all
    computations with gradients turned off.
    """

    def __init__(self, hook: BaseHook, model_name: str, **kwargs):
        """
        Arguments:
            hook: The wrapped hook which computes all losses and outputs.
            model_name: The name of the model that will be set to eval() mode.
        """

        super().__init__(**kwargs)
        self.hook = hook
        self.model_name = model_name

    def call(self, inputs, losses):
        """"""
        model = inputs[self.model_name]
        model.eval()
        with torch.no_grad():
            return self.hook(inputs, losses)



from typing import List, Union

import torch
from pytorch_metric_learning.distances import BatchedDistance, LpDistance
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f
from . import utils as l_u


def check_batch_sizes(s, t, mmd_type):
    if mmd_type == "quadratic":
        return
    is_list = c_f.is_list_or_tuple(s)
    if (is_list and any(s[i].shape != t[i].shape for i in range(len(s)))) or (
        not is_list and s.shape != t.shape
    ):
        raise ValueError(
            "For mmd_type 'linear', source and target must have the same batch size."
        )


class MMDLoss(torch.nn.Module):
    """
    Implementation of

    - [Learning Transferable Features with Deep Adaptation Networks](https://arxiv.org/abs/1502.02791)

    - [Deep Transfer Learning with Joint Adaptation Networks](https://arxiv.org/abs/1605.06636).
    """

    def __init__(
        self,
        kernel_scales: Union[float, torch.Tensor] = 1,
        mmd_type: str = "linear",
        dist_func=None,
        bandwidth=None,
    ):
        """
        Arguments:
            kernel_scales: The kernel bandwidth is scaled by this amount.
                If a tensor, then multiple kernel bandwidths are used.
            mmd_type: 'linear' or 'quadratic'. 'linear' uses the linear estimate of MK-MMD.
        """
        super().__init__()
        self.kernel_scales = kernel_scales
        self.dist_func = c_f.default(
            dist_func, LpDistance(normalize_embeddings=False, p=2, power=2)
        )
        self.bandwidth = bandwidth
        self.mmd_type = mmd_type
        if mmd_type == "linear":
            self.mmd_func = l_u.get_mmd_linear
        elif mmd_type == "quadratic":
            self.mmd_func = l_u.get_mmd_quadratic
        else:
            raise ValueError("mmd_type must be either linear or quadratic")

    # input can be embeddings or list of embeddings
    def forward(
        self,
        x: Union[torch.Tensor, List[torch.Tensor]],
        y: Union[torch.Tensor, List[torch.Tensor]],
    ) -> torch.Tensor:
        """
        Arguments:
            x: features or a list of features from one domain.
            y: features or a list of features from the other domain.
        Returns:
            MMD if the inputs are tensors, and Joint MMD (JMMD) if the inputs are lists of tensors.
        """
        check_batch_sizes(x, y, self.mmd_type)
        xx, yy, zz, scale = l_u.get_mmd_dist_mats(x, y, self.dist_func, self.bandwidth)
        if torch.is_tensor(self.kernel_scales):
            s = scale[0] if c_f.is_list_or_tuple(scale) else scale
            self.kernel_scales = pml_cf.to_device(self.kernel_scales, s, dtype=s.dtype)

        if c_f.is_list_or_tuple(scale):
            for i in range(len(scale)):
                scale[i] = scale[i] * self.kernel_scales
        else:
            scale = scale * self.kernel_scales

        return self.mmd_func(xx, yy, zz, scale)

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["mmd_type", "kernel_scales"])


class MMDBatchedLoss(MMDLoss):
    def __init__(self, batch_size=1024, **kwargs):
        super().__init__(**kwargs)
        if self.mmd_type != "quadratic":
            raise ValueError("mmd_type must be 'quadratic'")
        self.mmd_func = l_u.get_mmd_quadratic_batched
        self.dist_func = BatchedDistance(self.dist_func, batch_size=batch_size)

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: features from one domain.
            y: features from the other domain.
        Returns:
            MMD
        """
        if c_f.is_list_or_tuple(x) or c_f.is_list_or_tuple(y):
            raise TypeError("List of features not yet supported")
        check_batch_sizes(x, y, self.mmd_type)
        return self.mmd_func(x, y, self.dist_func, self.kernel_scales, self.bandwidth)



import logging

import ignite.distributed as idist
from ignite.contrib.handlers import ProgressBar
from ignite.engine import Events
from ignite.handlers import EarlyStopping
from ignite.utils import setup_logger

from ...utils import common_functions as c_f
from ...utils import exceptions
from ...validators import utils as val_utils
from ...weighters import get_multiple_loss_totals
from .dictionary_accumulator import DictionaryAccumulator


def get_validation_runner(
    collector,
    dataloaders,
    validator,
    val_hooks,
    logger,
):
    required_data = []
    for v in [validator, *val_hooks]:
        if v and hasattr(v, "required_data"):
            required_data = list(set(required_data + v.required_data))

    def run_validation(engine):
        epoch = engine.state.epoch
        collected_data = collect_from_dataloaders(collector, dataloaders, required_data)
        score = None
        if validator:
            score = val_utils.call_val_hook(validator, collected_data, epoch)
        for hook in val_hooks:
            val_utils.call_val_hook(hook, collected_data, epoch)
        if logger:
            logger.add_validation({"validator": validator}, epoch)
            logger.write(engine)
        return score

    return run_validation


def collect_from_dataloaders(collector, dataloaders, required_data):
    collected_data = {}

    for k in required_data:
        c_f.LOGGER.info(f"Collecting {k}")
        curr_dataloader = dataloaders[k]
        c_f.val_dataloader_checks(curr_dataloader)
        curr_dataset = curr_dataloader.dataset
        iterable = curr_dataloader.__iter__()
        curr_collected = accumulate_collector_output(collector, iterable, k)
        c_f.val_collected_data_checks(curr_collected, curr_dataset)
        collected_data[k] = curr_collected
        del iterable

    return collected_data


def step_lr_schedulers(lr_schedulers, scheduler_type):
    def handler(engine):
        lr_schedulers.step(scheduler_type)

    return handler


def auto_model(*args, **kwargs):
    def handler(model):
        return idist.auto_model(model, *args, **kwargs)

    return handler


def pbar_print_losses(pbar):
    def handler(engine):
        losses = get_multiple_loss_totals(engine.state.output)
        losses = [f"{k}={v:6.4f}" for k, v in losses.items()]
        pbar.pbar.unit = ", ".join(losses)

    return handler


def get_default_pbar():
    bar_format = "{desc}[{n_fmt}/{total_fmt}] {percentage:3.0f}%|{bar}|{unit} [{elapsed}<{remaining}]"
    return ProgressBar(persist=True, bar_format=bar_format)


def attach_pbar(engine, *_):
    pbar = get_default_pbar()
    pbar.attach(engine)
    return pbar


def accumulate_collector_output(collector, iterable, output_name):
    accumulator = DictionaryAccumulator()
    accumulator.attach(collector, output_name)
    collector.run(iterable)
    accumulator.detach(collector)
    output = collector.state.metrics[output_name]
    collector.state.output = {}
    collector.state.metrics = {}
    return output


def set_engine_logger(engine, name, level=logging.CRITICAL):
    engine.logger = setup_logger(name=name, level=level)
    return engine.logger


def register(engine, event, *args):
    for h in args:
        engine.add_event_handler(event, h)


def do_for_all_engines(cls, function, keys):
    output = {}
    for name in keys:
        output[name] = function(getattr(cls, name), name)
    return output


def set_loggers_and_pbars(cls, keys):
    do_for_all_engines(cls, set_engine_logger, keys)
    if cls.with_pbars:
        return do_for_all_engines(cls, attach_pbar, keys)


def resume_checks(trainer, validator):
    last_trainer_epoch = trainer.state.epoch
    if not validator:
        return
    last_validator_epoch = validator.epochs[-1]
    if last_trainer_epoch != last_validator_epoch:
        raise exceptions.ResumeCheckError(
            f"Last trainer epoch ({last_trainer_epoch}) does not equal last validator epoch ({last_validator_epoch})"
        )


def is_done(trainer, max_epochs):
    return trainer.state.epoch >= max_epochs


def zero_grad(adapter):
    def handler(engine):
        c_f.LOGGER.debug("zeroing grads")
        adapter.models.zero_grad()
        adapter.optimizers.zero_grad()

    return handler


def interval_condition(interval, max_epochs):
    condition = Events.EPOCH_COMPLETED(every=interval)
    if max_epochs % interval != 0:
        condition |= Events.EPOCH_COMPLETED(once=max_epochs)
    return condition


def early_stopper(**kwargs):
    def fn(trainer, score_function):
        return EarlyStopping(trainer=trainer, score_function=score_function, **kwargs)

    return fn



import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f
from .concat_softmax import ConcatSoftmax
from .utils import split_half


class SymNetsDomainLoss(torch.nn.Module):
    def __init__(self, half_idx):
        super().__init__()
        self.half_idx = half_idx
        self.softmax_fn = ConcatSoftmax()

    # x and y are the first and second halves of "p^st"
    def forward(self, x, y):
        x = self.softmax_fn(x, y)
        x = split_half(x, dim=1)[self.half_idx]
        x = torch.sum(x, dim=1)
        x[x == 0] += pml_cf.small_val(x.dtype)
        return -torch.mean(torch.log(x))

    def extra_repr(self):
        return c_f.extra_repr(self, ["half_idx"])



class IgniteEmptyLogger:
    def add_training(self, *args, **kwargs):
        def fn(*args, **kwargs):
            pass

        return fn

    def add_validation(self, *args, **kwargs):
        pass

    def write(self, *args, **kwargs):
        pass



from torch.hub import load_state_dict_from_url


def download_weights(model, url, pretrained, map_location=None, **kwargs):
    if pretrained:
        state_dict = load_state_dict_from_url(
            url, check_hash=True, map_location=map_location, **kwargs
        )
        if map_location:
            model.to(map_location)
        model.load_state_dict(state_dict)
    return model
